@@ -270,12 +270,11 @@ edit_gpgme_key (SeahorsePGPSource *psrc, gpgme_key_t key, SeahorseEditParm *parm
     
     g_assert (psrc && SEAHORSE_IS_PGP_SOURCE (psrc));
     
-    err = gpgme_data_new (&out);
-    g_return_val_if_fail (GPG_IS_OK (err), err);
+    out = gpgmex_data_new ();
     
     /* do edit callback, release data */
     err = gpgme_op_edit (psrc->gctx, key, seahorse_pgp_key_op_edit, parms, out);
-    gpgme_data_release (out);
+    gpgmex_data_release (out);
     
     return err;
 }
@@ -2084,7 +2083,7 @@ seahorse_pgp_key_op_del_uid (SeahorsePGPKey *pkey, const guint index)
     g_return_val_if_fail (SEAHORSE_IS_KEY (pkey), GPG_E (GPG_ERR_WRONG_KEY_USAGE));
     
     /* Check index range */
-    DEBUG_OPERATION(("Del UID: uid %i\n", index));
+    DEBUG_OPERATION(("Del UID: uid %u\n", index));
     g_return_val_if_fail (real_index >= 1 && real_index <= (seahorse_pgp_key_get_num_userids (pkey) + seahorse_pgp_key_get_num_photoids(pkey)), GPG_E (GPG_ERR_INV_VALUE));
 
     /* Make sure not revoked */
