@@ -32,6 +32,8 @@
 #include "seahorse-pgp-key-op.h"
 #include "seahorse-libdialogs.h"
 #include "seahorse-gpg-options.h"
+#include "seahorse-passphrase.h"
+#include "seahorse-pgp-operation.h"
 
 /* TODO: Verify properly that all keys we deal with are PGP keys */
 
@@ -151,7 +153,8 @@ static void                seahorse_pgp_source_stop             (SeahorseKeySour
 static guint               seahorse_pgp_source_get_state        (SeahorseKeySource *src);
 static SeahorseOperation*  seahorse_pgp_source_load             (SeahorseKeySource *src,
                                                                  SeahorseKeySourceLoad load,
-                                                                 const gchar *key);
+                                                                 GQuark keyid,
+                                                                 const gchar *match);
 static SeahorseOperation*  seahorse_pgp_source_import           (SeahorseKeySource *sksrc, 
                                                                  gpgme_data_t data);
 static SeahorseOperation*  seahorse_pgp_source_export           (SeahorseKeySource *sksrc, 
@@ -197,11 +200,11 @@ seahorse_pgp_source_class_init (SeahorsePGPSourceClass *klass)
  
     g_object_class_install_property (gobject_class, PROP_KEY_TYPE,
         g_param_spec_uint ("key-type", "Key Type", "Key type that originates from this key source.", 
-                           0, G_MAXUINT, SKEY_INVALID, G_PARAM_READABLE));
+                           0, G_MAXUINT, SKEY_UNKNOWN, G_PARAM_READABLE));
 
     g_object_class_install_property (gobject_class, PROP_LOCATION,
         g_param_spec_uint ("location", "Key Location", "Where the key is stored. See SeahorseKeyLoc", 
-                           0, G_MAXUINT, SKEY_LOC_UNKNOWN, G_PARAM_READABLE));    
+                           0, G_MAXUINT, SKEY_LOC_INVALID, G_PARAM_READABLE));    
 }
 
 /* init context, private vars, set prefs, connect signals */
@@ -259,7 +262,7 @@ seahorse_pgp_source_dispose (GObject *gobject)
     
     /* Clear out all operations */
     if (psrc->pv->operations) {
-        if(!seahorse_operation_is_done (SEAHORSE_OPERATION (psrc->pv->operations)))
+        if(seahorse_operation_is_running (SEAHORSE_OPERATION (psrc->pv->operations)))
             seahorse_operation_cancel (SEAHORSE_OPERATION (psrc->pv->operations));
         g_object_unref (psrc->pv->operations);
         psrc->pv->operations = NULL;
@@ -330,6 +333,7 @@ seahorse_pgp_source_get_property (GObject *object, guint prop_id, GValue *value,
 static void
 key_changed (SeahorseKey *skey, SeahorseKeyChange change, SeahorseKeySource *sksrc)
 {
+    SeahorsePGPKey *pkey;
     SeahorsePGPSource *psrc;
     SeahorseKeyEType etype;
     SeahorseLoadOperation *lop;
@@ -345,10 +349,14 @@ key_changed (SeahorseKey *skey, SeahorseKeyChange change, SeahorseKeySource *sks
     
     etype = seahorse_key_get_etype (skey);
     
+    g_return_if_fail (SEAHORSE_IS_PGP_KEY (skey));
+    pkey = SEAHORSE_PGP_KEY (skey);
+    g_return_if_fail (pkey->pubkey);
+    
     psrc = SEAHORSE_PGP_SOURCE (seahorse_key_get_source (skey));
     g_assert (SEAHORSE_IS_PGP_SOURCE (psrc));
     
-    patterns[0] = seahorse_key_get_keyid (skey);
+    patterns[0] = seahorse_pgp_key_get_id (pkey->pubkey, 0);
     patterns[1] = NULL;
     
     parts = 0;
@@ -381,23 +389,24 @@ key_destroyed (SeahorseKey *skey, SeahorseKeySource *sksrc)
 
 /* Remove the given key from the context */
 static void
-remove_key_from_context (const gchar *id, SeahorseKey *dummy, SeahorsePGPSource *psrc)
+remove_key_from_context (gpointer kt, SeahorseKey *dummy, SeahorsePGPSource *psrc)
 {
     /* This function gets called as a GHRFunc on the lctx->checks hashtable. */
+    GQuark keyid = GPOINTER_TO_UINT (kt);
     
     SeahorseKey *skey;
     
-    skey = seahorse_context_get_key (SCTX_APP (), SEAHORSE_KEY_SOURCE (psrc), id);
+    skey = seahorse_context_get_key (SCTX_APP (), SEAHORSE_KEY_SOURCE (psrc), keyid);
     if (skey != NULL)
         seahorse_context_remove_key (SCTX_APP (), skey);
 }
 
 static gboolean
-have_key_in_context (SeahorsePGPSource *psrc, const gchar *id, gboolean secret)
+have_key_in_context (SeahorsePGPSource *psrc, GQuark keyid, gboolean secret)
 {
     SeahorseKey *skey;
     g_assert (SEAHORSE_IS_PGP_SOURCE (psrc));
-    skey = seahorse_context_get_key (SCTX_APP (), SEAHORSE_KEY_SOURCE (psrc), id);
+    skey = seahorse_context_get_key (SCTX_APP (), SEAHORSE_KEY_SOURCE (psrc), keyid);
     return skey && (!secret || seahorse_key_get_etype (skey) == SKEY_PRIVATE);
 }
 
@@ -408,12 +417,15 @@ add_key_to_context (SeahorsePGPSource *psrc, gpgme_key_t key, gpgmex_photo_id_t
     SeahorsePGPKey *pkey = NULL;
     SeahorsePGPKey *prev;
     const gchar *id;
+    GQuark keyid;
     GList *l;
     
     id = seahorse_pgp_key_get_id (key, 0);
+    keyid = seahorse_pgp_key_get_cannonical_id (id);
+    g_return_val_if_fail (keyid, NULL);
     
     g_assert (SEAHORSE_IS_PGP_SOURCE (psrc));
-    prev = SEAHORSE_PGP_KEY (seahorse_context_get_key (SCTX_APP (), SEAHORSE_KEY_SOURCE (psrc), id));
+    prev = SEAHORSE_PGP_KEY (seahorse_context_get_key (SCTX_APP (), SEAHORSE_KEY_SOURCE (psrc), keyid));
     
     /* Check if we can just replace the key on the object */
     if (prev != NULL) {
@@ -490,9 +502,9 @@ scheduled_refresh (gpointer data)
 
     DEBUG_REFRESH ("scheduled refresh event ocurring now\n");
     cancel_scheduled_refresh (psrc);
-    seahorse_key_source_load_async (SEAHORSE_KEY_SOURCE (psrc), SKSRC_LOAD_ALL, NULL);    
+    seahorse_key_source_load_async (SEAHORSE_KEY_SOURCE (psrc), SKSRC_LOAD_ALL, 0, NULL);
     
-    return FALSE; /* don't run again */    
+    return FALSE; /* don't run again */
 }
 
 static gboolean
@@ -612,6 +624,7 @@ keyload_handler (SeahorseLoadOperation *lop)
     gpgme_key_t key;
     guint batch;
     const gchar *id;
+    GQuark keyid;
     gchar *t;
     
     g_assert (SEAHORSE_IS_LOAD_OPERATION (lop));
@@ -634,15 +647,22 @@ keyload_handler (SeahorseLoadOperation *lop)
         }
         
         id = seahorse_pgp_key_get_id (key, 0);
+        keyid = seahorse_pgp_key_get_cannonical_id (id);
+        
+        /* Invalid id from GPG ? */
+        if (!keyid) {
+            gpgmex_key_unref (key);
+            continue;
+        }
         
         /* During a refresh if only new or removed keys */
         if (lop->checks) {
 
             /* Make note that this key exists in keyring */
-            g_hash_table_remove (lop->checks, id);
+            g_hash_table_remove (lop->checks, GUINT_TO_POINTER (keyid));
 
             /* When not doing all keys and already have ... */
-            if (!lop->all && have_key_in_context (lop->psrc, id, lop->secret)) {
+            if (!lop->all && have_key_in_context (lop->psrc, keyid, lop->secret)) {
 
                 /* ... then just ignore */
                 gpgmex_key_unref (key);
@@ -672,7 +692,7 @@ keyload_handler (SeahorseLoadOperation *lop)
     
     /* TODO: We can use the GPGME progress to make this more accurate */
     t = g_strdup_printf (ngettext("Loaded %d key", "Loaded %d keys", lop->loaded), lop->loaded);
-    seahorse_operation_mark_progress (SEAHORSE_OPERATION (lop), t, 0, 0);
+    seahorse_operation_mark_progress (SEAHORSE_OPERATION (lop), t, 0.0);
     g_free (t);
     
     return TRUE; 
@@ -716,14 +736,14 @@ seahorse_load_operation_start (SeahorsePGPSource *psrc, const gchar **pattern, g
         /* This hashtable contains only allocated strings with no 
          * values. We allocate the strings in case a key goes away
          * while we're holding the ids in this table */
-        lop->checks = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
+        lop->checks = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, NULL);
         
         keys = seahorse_context_get_keys (SCTX_APP (), SEAHORSE_KEY_SOURCE (psrc));
         for (l = keys; l; l = g_list_next (l)) {
             skey = SEAHORSE_KEY (l->data);
             if (secret && seahorse_key_get_etype (skey) != SKEY_PRIVATE) 
                 continue;
-            g_hash_table_insert (lop->checks, g_strdup (seahorse_key_get_keyid (l->data)), NULL);
+            g_hash_table_insert (lop->checks, GUINT_TO_POINTER (seahorse_key_get_keyid (l->data)), NULL);
         }
         
         g_list_free (keys);
@@ -731,7 +751,7 @@ seahorse_load_operation_start (SeahorsePGPSource *psrc, const gchar **pattern, g
     }
     
     seahorse_operation_mark_start (SEAHORSE_OPERATION (lop));
-    seahorse_operation_mark_progress (SEAHORSE_OPERATION (lop), _("Loading Keys..."), 0, 0);
+    seahorse_operation_mark_progress (SEAHORSE_OPERATION (lop), _("Loading Keys..."), 0.0);
     
     if (all && !refresh) {
         /* Load keys at idle time */
@@ -746,13 +766,71 @@ seahorse_load_operation_start (SeahorsePGPSource *psrc, const gchar **pattern, g
     return lop;
 }    
 
+static void
+prepare_import_results (SeahorsePGPOperation *pop, SeahorsePGPSource *psrc)
+{
+    SeahorseKey *skey;
+    SeahorseLoadOperation *lop;
+    gpgme_import_result_t results;
+    gpgme_import_status_t import;
+    SeahorseKeySource *sksrc;
+    GList *keys = NULL;
+    const gchar **patterns = NULL;
+    GQuark keyid;
+    guint i;
+    
+    sksrc = SEAHORSE_KEY_SOURCE (psrc);
+
+    /* Figure out which keys were imported */
+    results = gpgme_op_import_result (pop->gctx);
+    if (results) {
+            
+        /* Dig out all the fingerprints for use as load patterns */
+        patterns = (const gchar**)g_new0(gchar*, results->considered + 1);
+        for (i = 0, import = results->imports; 
+             i < results->considered && import; 
+             import = import->next) {
+            if (GPG_IS_OK (import->result))
+                patterns[i++] = import->fpr;
+        }
+
+        /* Reload public keys */
+        lop = seahorse_load_operation_start (psrc, patterns, LOAD_FULL, 
+                                             FALSE, FALSE, TRUE);
+        g_object_unref (lop);
+
+        /* Reload secret keys */
+        lop = seahorse_load_operation_start (psrc, patterns, LOAD_FULL,
+                                             TRUE, FALSE, TRUE);
+        g_object_unref (lop);
+            
+        g_free (patterns);
+            
+        /* Now get a list of the new keys */
+        for (import = results->imports; import; import = import->next) {
+            if (!GPG_IS_OK (import->result))
+                continue;
+            
+            keyid = seahorse_pgp_key_get_cannonical_id (import->fpr);
+            if (keyid) {
+                skey = seahorse_context_get_key (SCTX_APP (), sksrc, keyid);
+                if (skey != NULL)
+                    keys = g_list_prepend (keys, skey);
+            }
+        }
+    }
+    
+    seahorse_operation_mark_result (SEAHORSE_OPERATION (pop), keys, 
+                                    (GDestroyNotify)g_list_free);
+}
+
 /* --------------------------------------------------------------------------
  * METHODS
  */
 
 static SeahorseOperation*
 seahorse_pgp_source_load (SeahorseKeySource *src, SeahorseKeySourceLoad load,
-                          const gchar *match)
+                          GQuark keyid, const gchar *match)
 {
     SeahorsePGPSource *psrc;
     SeahorseLoadOperation *lop;
@@ -773,6 +851,9 @@ seahorse_pgp_source_load (SeahorseKeySource *src, SeahorseKeySourceLoad load,
     if (ref || all) {
         ref = TRUE;
         match = NULL;
+    } else if (load == SKSRC_LOAD_KEY) {
+        g_return_val_if_fail (keyid, NULL);
+        match = seahorse_key_get_rawid (keyid);
     }
     
     patterns[0] = match;
@@ -806,7 +887,7 @@ seahorse_pgp_source_stop (SeahorseKeySource *src)
     g_assert (SEAHORSE_IS_KEY_SOURCE (src));
     psrc = SEAHORSE_PGP_SOURCE (src);
     
-    if(!seahorse_operation_is_done (SEAHORSE_OPERATION (psrc->pv->operations)))
+    if(seahorse_operation_is_running (SEAHORSE_OPERATION (psrc->pv->operations)))
         seahorse_operation_cancel (SEAHORSE_OPERATION (psrc->pv->operations));
 }
 
@@ -818,160 +899,108 @@ seahorse_pgp_source_get_state (SeahorseKeySource *src)
     g_assert (SEAHORSE_IS_KEY_SOURCE (src));
     psrc = SEAHORSE_PGP_SOURCE (src);
     
-    return seahorse_operation_is_done (SEAHORSE_OPERATION (psrc->pv->operations)) ? 
-                                       SKSRC_LOADING : 0;
+    return seahorse_operation_is_running (SEAHORSE_OPERATION (psrc->pv->operations)) ? 
+                                          SKSRC_LOADING : 0;
 }
 
 static SeahorseOperation* 
 seahorse_pgp_source_import (SeahorseKeySource *sksrc, gpgme_data_t data)
 {
-    SeahorseLoadOperation *lop;
-    SeahorseOperation *operation;
+    SeahorsePGPOperation *pop;
     SeahorsePGPSource *psrc;
-    gpgme_import_result_t results;
-    gpgme_import_status_t import;
-    SeahorseKey *skey;
     gpgme_error_t gerr;
-    gpgme_ctx_t new_ctx;
-    GList *keys = NULL;
-    GError *err = NULL;
-    const gchar **patterns = NULL;
-	guint i;
     
     g_return_val_if_fail (SEAHORSE_IS_PGP_SOURCE (sksrc), NULL);
     psrc = SEAHORSE_PGP_SOURCE (sksrc);
     
-    /* Eventually this should probably be asynchronous, but for now
-     * we leave it as a sync operation for simplicity. */
+    pop = seahorse_pgp_operation_new (_("Importing Keys"));
+    g_return_val_if_fail (pop != NULL, NULL);
     
-    new_ctx = seahorse_pgp_source_new_context ();
-    g_return_val_if_fail (new_ctx != NULL, NULL);
+    gpgme_data_seek (data, 0, SEEK_SET);
+    gerr = gpgme_op_import_start (pop->gctx, data);
     
-    operation = g_object_new (SEAHORSE_TYPE_OPERATION, NULL);
-    seahorse_operation_mark_start (operation);
+    g_signal_connect (pop, "results", G_CALLBACK (prepare_import_results), psrc);
+    g_object_set_data_full (G_OBJECT (pop), "source-data", data, 
+                            (GDestroyNotify)gpgmex_data_release);
     
-    gerr = gpgme_op_import (new_ctx, data);
-    if (GPG_IS_OK (gerr)) {
-
-        /* Figure out which keys were imported */
-        results = gpgme_op_import_result (new_ctx);
-        if (results) {
-            
-            /* Dig out all the fingerprints for use as load patterns */
-            patterns = (const gchar**)g_new0(gchar*, results->considered + 1);
-            for (i = 0, import = results->imports; 
-				 i < results->considered && import; 
-				 import = import->next) {
-                if (GPG_IS_OK (import->result))
-                    patterns[i++] = import->fpr;
-            }
-
-            /* Reload public keys */
-            lop = seahorse_load_operation_start (psrc, patterns, LOAD_FULL, 
-                                                 FALSE, FALSE, TRUE);
-            g_object_unref (lop);
-
-            /* Reload secret keys */
-            lop = seahorse_load_operation_start (psrc, patterns, LOAD_FULL,
-                                                 TRUE, FALSE, TRUE);
-            g_object_unref (lop);
-            
-            g_free (patterns);
-            
-            /* Now get a list of the new keys */
-            for (import = results->imports; import; import = import->next) {
-                if (!GPG_IS_OK (import->result))
-                    continue;
-                    
-                skey = seahorse_context_get_key (SCTX_APP (), sksrc, import->fpr);
-                if (skey != NULL)
-                    keys = g_list_prepend (keys, skey);
-            }
-        }        
-        
-        seahorse_operation_mark_result (operation, keys, (GDestroyNotify)g_list_free);
-        seahorse_operation_mark_done (operation, FALSE, NULL);
-        
-    } else {
-       
-        /* Marks the operation as failed */
-        seahorse_util_gpgme_to_error (gerr, &err);
-        seahorse_operation_mark_done (operation, FALSE, err);
-    }
-
-    gpgme_release (new_ctx);
-    return operation;                    
+    /* Couldn't start import */
+    if (!GPG_IS_OK (gerr))
+        seahorse_pgp_operation_mark_failed (pop, gerr);
+    
+    return SEAHORSE_OPERATION (pop);
 }
 
 static SeahorseOperation* 
 seahorse_pgp_source_export (SeahorseKeySource *sksrc, GList *keys, 
                             gboolean complete, gpgme_data_t data)
 {
-    SeahorseOperation *operation;
+    SeahorsePGPOperation *pop;
     SeahorsePGPSource *psrc;
     SeahorsePGPKey *pkey;
     SeahorseKey *skey;
-    gpgme_error_t gerr;
-    gpgme_ctx_t new_ctx;
-    GError *err = NULL;
+    gpgme_error_t gerr = GPG_OK;
+    GArray *akeyids; 
+    gchar **keyids;
+    const gchar *keyid;
     GList *l;
     
     g_return_val_if_fail (SEAHORSE_IS_PGP_SOURCE (sksrc), NULL);
     psrc = SEAHORSE_PGP_SOURCE (sksrc);
-    
-    /* Eventually this should probably be asynchronous, but for now
-     * we leave it as a sync operation for simplicity. */
 
-    operation = g_object_new (SEAHORSE_TYPE_OPERATION, NULL);
-    seahorse_operation_mark_start (operation);
-     
+    pop = seahorse_pgp_operation_new (_("Exporting Keys"));
+    g_return_val_if_fail (pop != NULL, NULL);
+
     if (data == NULL) {
         
         /* Tie to operation when we create our own (auto-free) */
-        gerr = gpgme_data_new (&data);
-        g_return_val_if_fail (GPG_IS_OK (gerr), NULL);
-        seahorse_operation_mark_result (operation, data, 
-                                        (GDestroyNotify)gpgme_data_release);
+        data = gpgmex_data_new ();
+        seahorse_operation_mark_result (SEAHORSE_OPERATION (pop), data, 
+                                        (GDestroyNotify)gpgmex_data_release);
     } else {
        
         /* Don't free when people pass us their own data */
-        seahorse_operation_mark_result (operation, data, NULL);
+        seahorse_operation_mark_result (SEAHORSE_OPERATION (pop), data, NULL);
     }
-     
-    new_ctx = seahorse_pgp_source_new_context ();
-    g_return_val_if_fail (new_ctx != NULL, NULL);
-
-    gpgme_set_armor (new_ctx, TRUE);
-    gpgme_set_textmode (new_ctx, TRUE);
+    
+    gpgme_set_armor (pop->gctx, TRUE);
+    gpgme_set_textmode (pop->gctx, TRUE);
 
+    /* Build list of keyids for export */
+    akeyids = g_array_new (TRUE, TRUE, sizeof (gchar*));
+    
     for (l = keys; l != NULL; l = g_list_next (l)) {
-       
+        
         g_return_val_if_fail (SEAHORSE_IS_PGP_KEY (l->data), NULL);
         pkey = SEAHORSE_PGP_KEY (l->data);
         
         skey = SEAHORSE_KEY (l->data);       
         g_return_val_if_fail (skey->sksrc == sksrc, NULL);
         
-        gerr = gpgme_op_export (new_ctx, skey->keyid, 0, data);
-
-        if (!GPG_IS_OK (gerr))
-            break;
+        /* Building list */
+        keyid = seahorse_pgp_key_get_id (pkey->pubkey, 0);
+        g_array_append_val (akeyids, keyid);
         
+        /* TODO: Exporting of secret keys is synchronous */
         if (complete && skey->etype == SKEY_PRIVATE) {
             
-            gerr = gpgmex_op_export_secret (new_ctx, 
+            gerr = gpgmex_op_export_secret (pop->gctx, 
                         seahorse_pgp_key_get_id (pkey->seckey, 0), data);
             
             if (!GPG_IS_OK (gerr))
                 break;
         }
     }
-
-    if (!GPG_IS_OK (gerr)) 
-        seahorse_util_gpgme_to_error (gerr, &err);
-                                 
-    seahorse_operation_mark_done (operation, FALSE, err);
-    return operation;    
+    
+    keyids = (gchar**)g_array_free (akeyids, FALSE);
+    
+    if (GPG_IS_OK (gerr))
+        gerr = gpgme_op_export_ext_start (pop->gctx, (const char**)keyids, 0, data);
+    
+    if (!GPG_IS_OK (gerr))
+        seahorse_pgp_operation_mark_failed (pop, gerr);
+    
+    g_free (keyids);
+    return SEAHORSE_OPERATION (pop);
 }
 
 static gboolean            
@@ -995,7 +1024,7 @@ seahorse_pgp_source_remove (SeahorseKeySource *sksrc, SeahorseKey *skey,
         return FALSE;
     }
     
-    seahorse_key_source_load_async (sksrc, SKSRC_LOAD_NEW, NULL);    
+    seahorse_key_source_load_async (sksrc, SKSRC_LOAD_NEW, 0, NULL);
     return TRUE;
 }
 
