@@ -30,6 +30,7 @@
 #include "seahorse-server-source.h"
 #include "seahorse-gconf.h"
 #include "seahorse-pgp-key.h"
+#include "seahorse-gtkstock.h"
 
 /* From seahorse_preferences_cache.c */
 void seahorse_prefs_cache (SeahorseWidget *widget);
@@ -425,6 +426,11 @@ setup_keyservers (SeahorseWidget *swidget)
     skc = seahorse_keyserver_control_new (PUBLISH_TO_KEY, _("None: Don't publish keys"));
     gtk_container_add (GTK_CONTAINER (w), GTK_WIDGET (skc));
     gtk_widget_show_all (w);
+    
+    w = glade_xml_get_widget(swidget->xml, "auto_sync");
+    g_return_if_fail (w != NULL);
+    
+    seahorse_check_button_gconf_attach (GTK_CHECK_BUTTON(w), AUTOSYNC_KEY);
 }
 
 #endif /* WITH_KEYSERVER */
@@ -436,11 +442,13 @@ setup_keyservers (SeahorseWidget *swidget)
 static void
 setup_sharing (SeahorseWidget *swidget)
 {
-    GtkWidget *widget;
-    widget = glade_xml_get_widget (swidget->xml, "enable-sharing-placeholder");
-    gtk_container_add (GTK_CONTAINER (widget),
-        seahorse_check_button_control_new (_("_Share my keys with others on my network"), KEYSHARING_KEY));
-    gtk_widget_show_all (widget);
+    GtkWidget *widget = seahorse_widget_get_widget (swidget, "enable-sharing");
+    g_return_if_fail (widget != NULL);
+    seahorse_check_button_gconf_attach (GTK_CHECK_BUTTON (widget), KEYSHARING_KEY);
+    
+    widget = seahorse_widget_get_widget (swidget, "key-sharing-image");
+    g_return_if_fail (widget != NULL);
+    gtk_image_set_from_stock (GTK_IMAGE (widget), SEAHORSE_STOCK_SHARING, GTK_ICON_SIZE_DIALOG);
 }
 
 #endif /* WITH_SHARING */
@@ -448,18 +456,18 @@ setup_sharing (SeahorseWidget *swidget)
 /* -------------------------------------------------------------------------- */
 
 static void
-default_key_changed (SeahorseDefaultKeyControl *sdkc, gpointer *data)
+default_key_changed (GtkOptionMenu *combo, gpointer *data)
 {
-    const gchar *id = seahorse_default_key_control_active_id (sdkc);
-    seahorse_gconf_set_string (DEFAULT_KEY, id == NULL ? "" : id);    
+    GQuark keyid = seahorse_combo_keys_get_active_id (combo);
+    seahorse_gconf_set_string (SEAHORSE_DEFAULT_KEY, keyid == 0 ? "" : g_quark_to_string (keyid));
 }
 
 static void
 gconf_notification (GConfClient *gclient, guint id, GConfEntry *entry, 
-                    SeahorseDefaultKeyControl *sdkc)
+                    GtkOptionMenu *combo)
 {
-    const gchar *key_id = gconf_value_get_string (gconf_entry_get_value (entry));
-    seahorse_default_key_control_select_id (sdkc, key_id);
+    GQuark keyid = g_quark_from_string (gconf_value_get_string (gconf_entry_get_value (entry)));
+    seahorse_combo_keys_set_active_id (combo, keyid);
 }
 
 static void
@@ -482,36 +490,34 @@ seahorse_prefs_new ()
     SeahorseKeyset *skset;
     SeahorseWidget *swidget;
     GtkWidget *widget;
-    SeahorseDefaultKeyControl *sdkc;
     guint gconf_id;
     
     swidget = seahorse_widget_new ("prefs");
     
-    widget = glade_xml_get_widget (swidget->xml, "modes");
-    gtk_container_add (GTK_CONTAINER (widget),
-        seahorse_check_button_control_new (_("_Encrypt to Self"), ENCRYPTSELF_KEY));
-    gtk_container_add (GTK_CONTAINER (widget),
-        seahorse_check_button_control_new (_("_Armor Encode Files"), ARMOR_KEY));
-    gtk_widget_show_all (widget);
+    widget = glade_xml_get_widget (swidget->xml, "encrypt-self");
+    seahorse_check_button_gconf_attach (GTK_CHECK_BUTTON (widget), ENCRYPTSELF_KEY);
+    
+    widget = glade_xml_get_widget (swidget->xml, "armor-encode");
+    seahorse_check_button_gconf_attach (GTK_CHECK_BUTTON (widget), ARMOR_KEY);
 
-    widget = glade_xml_get_widget (swidget->xml, "default_key");
+    widget = glade_xml_get_widget (swidget->xml, "signer-select");
+    g_return_val_if_fail (widget != NULL, NULL);
 
     skset = seahorse_keyset_new (SKEY_PGP, 
                                  SKEY_PRIVATE, 
                                  SKEY_LOC_LOCAL, 
-                                 0, 
+                                 SKEY_FLAG_CAN_SIGN, 
                                  SKEY_FLAG_EXPIRED | SKEY_FLAG_REVOKED | SKEY_FLAG_DISABLED);
-    sdkc = seahorse_default_key_control_new (skset, _("None. Prompt for a key."));
+    seahorse_combo_keys_attach (GTK_OPTION_MENU (widget), skset, _("None. Prompt for a key."));
     g_object_unref (skset);
     
-    gtk_container_add (GTK_CONTAINER (widget), GTK_WIDGET (sdkc));
-    gtk_widget_show_all (widget);
+    seahorse_combo_keys_set_active_id (GTK_OPTION_MENU (widget), 
+                                       g_quark_from_string (seahorse_gconf_get_string (SEAHORSE_DEFAULT_KEY)));
+    g_signal_connect (widget, "changed", G_CALLBACK (default_key_changed), NULL);
 
-    seahorse_default_key_control_select_id (sdkc, seahorse_gconf_get_string (DEFAULT_KEY));    
-    g_signal_connect (sdkc, "changed", G_CALLBACK (default_key_changed), NULL);
-
-    gconf_id = seahorse_gconf_notify (DEFAULT_KEY, (GConfClientNotifyFunc)gconf_notification, sdkc);
-    g_signal_connect (sdkc, "destroy", G_CALLBACK (remove_gconf_notification), GINT_TO_POINTER (gconf_id));
+    gconf_id = seahorse_gconf_notify (SEAHORSE_DEFAULT_KEY, 
+                                      (GConfClientNotifyFunc)gconf_notification, GTK_OPTION_MENU (widget));
+    g_signal_connect (widget, "destroy", G_CALLBACK (remove_gconf_notification), GINT_TO_POINTER (gconf_id));
     
 #ifdef WITH_AGENT   
     seahorse_prefs_cache (swidget);
