@@ -39,7 +39,8 @@ enum {
     PROP_FINGERPRINT,
     PROP_VALIDITY,
     PROP_TRUST,
-    PROP_EXPIRES
+    PROP_EXPIRES,
+    PROP_LENGTH
 };
 
 struct _SeahorseSSHKeyPrivate {
@@ -77,7 +78,7 @@ parse_ssh_algo (gchar *type)
         return SSH_ALGO_RSA1;
     if (strstr (type, "rsa"))
         return SSH_ALGO_RSA;
-    if (strstr (type, "dsa"))
+    if (strstr (type, "dsa") || strstr (type, "dss"))
         return SSH_ALGO_DSA;
     return SSH_ALGO_UNK;
 }
@@ -116,25 +117,28 @@ changed_key (SeahorseSSHKey *skey)
     
     /* Now start setting the main SeahorseKey fields */
     key->ktype = SKEY_SSH;
+    key->keyid = 0;
     
     if (!skey->keydata || !skey->keydata->keyid) {
         
-        key->keyid = "UNKNOWN ";
-        key->location = SKEY_LOC_UNKNOWN;
-        key->etype = SKEY_INVALID;
+        key->location = SKEY_LOC_INVALID;
+        key->etype = SKEY_ETYPE_NONE;
         key->loaded = SKEY_INFO_NONE;
         key->flags = SKEY_FLAG_DISABLED;
         
     } else {
     
         /* The key id */
-        key->keyid = skey->keydata->keyid ? skey->keydata->keyid : "UNKNOWN ";
+        key->keyid = seahorse_ssh_key_get_cannonical_id (skey->keydata->keyid);
         key->location = SKEY_LOC_LOCAL;
         key->etype = SKEY_PRIVATE;
         key->loaded = SKEY_INFO_COMPLETE;
         key->flags = SKEY_FLAG_TRUSTED;
     }
     
+    if (!key->keyid)
+        key->keyid = g_quark_from_string (SKEY_SSH_STR ":UNKNOWN ");
+    
     seahorse_key_changed (key, SKEY_CHANGE_ALL);
 }
 
@@ -200,6 +204,9 @@ seahorse_ssh_key_get_property (GObject *object, guint prop_id,
     case PROP_EXPIRES:
         g_value_set_ulong (value, 0);
         break;
+    case PROP_LENGTH:
+        g_value_set_uint (value, skey->keydata ? skey->keydata->length : 0);
+        break;
     }
 }
 
@@ -288,6 +295,10 @@ seahorse_ssh_key_class_init (SeahorseSSHKeyClass *klass)
     g_object_class_install_property (gobject_class, PROP_EXPIRES,
         g_param_spec_ulong ("expires", "Expires On", "Date this key expires on",
                            0, G_MAXULONG, 0, G_PARAM_READABLE));
+
+    g_object_class_install_property (gobject_class, PROP_LENGTH,
+        g_param_spec_uint ("length", "Length of", "The length of this key",
+                           0, G_MAXUINT, 0, G_PARAM_READABLE));
 }
 
 /* -----------------------------------------------------------------------------
@@ -307,8 +318,15 @@ seahorse_ssh_key_new (SeahorseKeySource *sksrc, SeahorseSSHKeyData *data)
     return skey;
 }
 
-const gchar*
+guint 
 seahorse_ssh_key_get_algo (SeahorseSSHKey *skey)
+{
+    g_return_val_if_fail (SEAHORSE_IS_SSH_KEY (skey), SSH_ALGO_UNK);
+    return skey->keydata->algo;
+}
+
+const gchar*
+seahorse_ssh_key_get_algo_str (SeahorseSSHKey *skey)
 {
     g_return_val_if_fail (SEAHORSE_IS_SSH_KEY (skey), "");
     
@@ -343,6 +361,36 @@ seahorse_ssh_key_get_filename (SeahorseSSHKey *skey, gboolean private)
     return private ? skey->keydata->filename : skey->keydata->filepub;
 }
 
+GQuark
+seahorse_ssh_key_get_cannonical_id (const gchar *id)
+{
+    static guint x = strlen (SKEY_SSH_STR);
+    gchar *p, *hex;
+    guint l, off = 0;
+    GQuark ret;
+
+    hex = g_strdup_printf ("%s:%s", SKEY_SSH_STR, id);
+    
+    /* Strip out all non alpha numeric chars */
+    for(p = hex + x + 1; *p; p++) {
+        if (g_ascii_isalnum(*p))
+            p[0 - off] = p[0];
+        else
+            off++;
+    }
+    p[0 - off] = 0;
+    
+    /* Shorten where necessary */
+    l = strlen (hex);
+    if (l > (8 + x + 1))
+        hex[8 + x + 1] = 0;
+    
+    ret = g_quark_from_string (hex);
+    g_free (hex);
+    
+    return ret;
+}
+
 /* -----------------------------------------------------------------------------
  * SSH KEY DATA 
  */
