@@ -37,10 +37,15 @@
 #include "seahorse-key-dialogs.h"
 #include "seahorse-pgp-key-op.h"
 #include "seahorse-key-widget.h"
-#include "seahorse-op.h"
 #include "seahorse-gpg-options.h"
 #include "seahorse-gconf.h"
 #include "seahorse-gtkstock.h"
+#include "seahorse-key-source.h"
+#include "seahorse-vfs-data.h"
+
+#ifdef WITH_KEYSERVER
+#include "seahorse-keyserver-sync.h"
+#endif
 
 #ifdef WITH_SSH
 #include "seahorse-ssh-key.h"
@@ -55,7 +60,7 @@ enum KeyManagerTabs {
 
 SeahorseKeyPredicate pred_public = {
     0,                  /* ktype, set later */
-    NULL,               /* keyid */
+    0,                  /* keyid */
     SKEY_LOC_LOCAL,     /* location */
     SKEY_PUBLIC,        /* etype */
     0,                  /* flags */
@@ -65,7 +70,7 @@ SeahorseKeyPredicate pred_public = {
 
 SeahorseKeyPredicate pred_trusted = {
     0,                  /* ktype, set later*/
-    NULL,               /* keyid */
+    0,                  /* keyid */
     SKEY_LOC_LOCAL,     /* location */
     SKEY_PUBLIC,        /* etype */
     SKEY_FLAG_TRUSTED,  /* flags */
@@ -75,7 +80,7 @@ SeahorseKeyPredicate pred_trusted = {
 
 SeahorseKeyPredicate pred_private = {
     0,                  /* ktype, set later */
-    NULL,               /* keyid */
+    0,                  /* keyid */
     SKEY_LOC_LOCAL,     /* location */
     SKEY_PRIVATE,       /* etype */
     0,                  /* flags */
@@ -205,19 +210,86 @@ new_button_clicked (GtkWidget *widget, SeahorseWidget *swidget)
 	seahorse_generate_druid_show ();
 }
 
+static void
+imported_keys (SeahorseOperation *op, SeahorseWidget *swidget)
+{
+    GError *err = NULL;
+    GList *keys;
+    guint nkeys;
+    
+    /* This key list is freed with the operation */
+    keys = (GList*)seahorse_operation_get_result (op);
+    nkeys = g_list_length (keys);
+    
+    if (seahorse_operation_is_successful (op)) {
+        set_numbered_status (swidget, _("Imported %d key"), 
+                                      _("Imported %d keys"), nkeys);
+    } else  {
+        seahorse_operation_steal_error (op, &err);
+        seahorse_util_handle_error (err, _("Couldn't import keys"));
+    }
+}
+
+static void
+import_files (SeahorseWidget *swidget, const gchar **uris)
+{
+    SeahorseKeySource *sksrc;
+    SeahorseOperation *op;
+    gpgme_data_t data;
+    GError *err = NULL;
+    
+    data = seahorse_vfs_data_read_multi (uris, &err);
+    if (!data) {
+        seahorse_util_handle_error (err, _("Couldn't import keys"));
+        return;
+    }
+
+    /* TODO: This needs work once we get more key types */
+    sksrc = seahorse_context_find_key_source (SCTX_APP (), SKEY_PGP, SKEY_LOC_LOCAL);
+    g_return_if_fail (sksrc && SEAHORSE_IS_PGP_SOURCE (sksrc));
+        
+    /* data is freed here */
+    op = seahorse_key_source_import (sksrc, data);
+    g_return_if_fail (op != NULL);
+
+    g_signal_connect (op, "done", G_CALLBACK (imported_keys), swidget);
+    seahorse_progress_show (op, _("Importing Keys"), TRUE);
+    
+    /* A running operation refs itself */
+    g_object_unref (op);
+}
+
+static void
+import_text (SeahorseWidget *swidget, const gchar *text)
+{
+    SeahorseKeySource *sksrc;
+    SeahorseOperation *op;
+    gpgme_data_t data;
+    
+    data = gpgmex_data_new_from_mem (text, strlen (text), TRUE);
+    g_return_if_fail (data != NULL);
+
+    /* TODO: This needs work once we get more key types */
+    sksrc = seahorse_context_find_key_source (SCTX_APP (), SKEY_PGP, SKEY_LOC_LOCAL);
+    g_return_if_fail (sksrc && SEAHORSE_IS_PGP_SOURCE (sksrc));
+    
+    /* data is freed here */
+    op = seahorse_key_source_import (sksrc, data);
+    g_return_if_fail (op != NULL);
+    
+    g_signal_connect (op, "done", G_CALLBACK (imported_keys), swidget);
+    seahorse_progress_show (op, _("Importing Keys"), TRUE);
+    
+    /* A running operation refs itself */
+    g_object_unref (op);
+}
+
 /* Loads import dialog */
 static void
 import_activate (GtkWidget *widget, SeahorseWidget *swidget)
 {
-    SeahorseKeySource *sksrc;
     GtkWidget *dialog;
     char* uri = NULL;
-    GError *err = NULL;
-    gint keys;
-    
-    /* TODO: This needs work once we get more key types */
-    sksrc = seahorse_context_find_key_source (SCTX_APP (), SKEY_PGP, SKEY_LOC_LOCAL);
-    g_return_if_fail (sksrc && SEAHORSE_IS_PGP_SOURCE (sksrc));
     
     dialog = seahorse_util_chooser_open_new (_("Import Key"), 
                 GTK_WINDOW(glade_xml_get_widget (swidget->xml, "key-manager")));
@@ -225,38 +297,17 @@ import_activate (GtkWidget *widget, SeahorseWidget *swidget)
 
     uri = seahorse_util_chooser_open_prompt (dialog);
     if(uri) {
-        
-        keys = seahorse_op_import_file (SEAHORSE_PGP_SOURCE (sksrc), uri, &err);
-        
-        if (err != NULL)
-            seahorse_util_handle_error (err, _("Couldn't import keys from \"%s\""), 
-                seahorse_util_uri_get_last (uri));
-        else 
-            set_numbered_status (swidget, _("Imported %d key"), 
-                                          _("Imported %d keys"), keys);
-
-        g_free (uri);
+        const gchar *uris[] = { uri, NULL };
+        import_files (swidget, uris);
     }
+    g_free (uri);
 }
 
 /* Callback for pasting from clipboard */
 static void
 clipboard_received (GtkClipboard *board, const gchar *text, SeahorseWidget *swidget)
 {
-    SeahorseKeySource *sksrc;
-    GError *err = NULL;
-    gint keys;
-    
-    sksrc = seahorse_context_find_key_source (SCTX_APP (), SKEY_PGP, SKEY_LOC_LOCAL);
-    g_return_if_fail (sksrc && SEAHORSE_IS_PGP_SOURCE (sksrc));
- 
-    keys = seahorse_op_import_text (SEAHORSE_PGP_SOURCE (sksrc), text, &err);
- 
-    if (err != NULL)
-        seahorse_util_handle_error (err, _("Couldn't import keys from clipboard"));
-    else
-        set_numbered_status (swidget, _("Imported %d key"), 
-                                      _("Imported %d keys"), keys);
+    import_text (swidget, text);
 }
 
 /* Pastes key from keyboard */
@@ -272,14 +323,44 @@ paste_activate (GtkWidget *widget, SeahorseWidget *swidget)
          (GtkClipboardTextReceivedFunc)clipboard_received, swidget);
 }
 
-/* Copies key to clipboard */
 static void
-copy_activate (GtkWidget *widget, SeahorseWidget *swidget)
+copy_done (SeahorseOperation *op, SeahorseWidget *swidget)
 {
     GdkAtom atom;
     GtkClipboard *board;
     gchar *text;
     GError *err = NULL;
+    gpgme_data_t data;
+    guint num;
+    
+    if (!seahorse_operation_is_successful (op)) {
+        seahorse_operation_steal_error (op, &err);
+        seahorse_util_handle_error (err, _("Couldn't retrieve data from key server"));
+    }
+    
+    data = (gpgme_data_t)seahorse_operation_get_result (op);
+    g_return_if_fail (data != NULL);
+    
+    text = seahorse_util_write_data_to_text (data, FALSE);
+    g_return_if_fail (text != NULL);
+    
+    atom = gdk_atom_intern ("CLIPBOARD", FALSE);
+    board = gtk_clipboard_get (atom);
+    gtk_clipboard_set_text (board, text, strlen (text));
+    g_free (text);
+
+    num = GPOINTER_TO_INT (g_object_get_data (G_OBJECT (op), "num-keys"));
+    if (num > 0) {
+        set_numbered_status (swidget, _("Copied %d key"), 
+                                      _("Copied %d keys"), num);
+    }
+}
+
+/* Copies key to clipboard */
+static void
+copy_activate (GtkWidget *widget, SeahorseWidget *swidget)
+{
+    SeahorseOperation *op;
     GList *keys;
     guint num;
   
@@ -288,21 +369,22 @@ copy_activate (GtkWidget *widget, SeahorseWidget *swidget)
     
     if (num == 0)
         return;
-               
-    text = seahorse_op_export_text (keys, FALSE, &err);
-    g_list_free (keys);
     
-    if (text == NULL)
-        seahorse_util_handle_error (err, _("Couldn't export key to clipboard"));
-    else {
-        atom = gdk_atom_intern ("CLIPBOARD", FALSE);
-        board = gtk_clipboard_get (atom);
-        gtk_clipboard_set_text (board, text, strlen (text));
-        g_free (text);
-
-        set_numbered_status (swidget, _("Copied %d key"), 
-                                      _("Copied %d keys"), num);
+    op = seahorse_key_source_export_keys (keys, NULL);
+    g_return_if_fail (op != NULL);
+    
+    g_object_set_data (G_OBJECT (op), "num-keys", GINT_TO_POINTER (num));
+        
+    if (seahorse_operation_is_running (op)) {
+        seahorse_progress_show (op, _("Retrieving keys"), TRUE);
+        g_signal_connect (op, "done", G_CALLBACK (copy_done), swidget);
+    } else {
+        copy_done (op, swidget);
     }
+        
+    /* Running operation refs itself */
+    g_object_unref (op);
+    g_list_free (keys);
 }
 
 /* Loads key properties if a key is selected */
@@ -318,9 +400,11 @@ properties_activate (GtkWidget *widget, SeahorseWidget *swidget)
 static void
 export_activate (GtkWidget *widget, SeahorseWidget *swidget)
 {
+    SeahorseOperation *op;
     GtkWidget *dialog;
     gchar* uri = NULL;
     GError *err = NULL;
+    gpgme_data_t data;
     GList *keys;
 
     keys = get_selected_keys (swidget);
@@ -334,16 +418,25 @@ export_activate (GtkWidget *widget, SeahorseWidget *swidget)
      
     uri = seahorse_util_chooser_save_prompt (dialog);
     if(uri) {
-
-        seahorse_op_export_file (keys, FALSE, uri, &err); 
-		        
-        if (err != NULL)
+        data = seahorse_vfs_data_create (uri, SEAHORSE_VFS_WRITE, &err);
+        if (data) {
+    
+            op = seahorse_key_source_export_keys (keys, data);
+            g_return_if_fail (op != NULL);
+        
+            seahorse_operation_wait (op);
+            gpgmex_data_release (data);
+    
+            if (!seahorse_operation_is_successful (op)) 
+                seahorse_operation_steal_error (op, &err);
+        }
+        
+        if (err) 
             seahorse_util_handle_error (err, _("Couldn't export key to \"%s\""),
-                    seahorse_util_uri_get_last (uri));
-                    
-        g_free (uri);
+                                        seahorse_util_uri_get_last (uri));
     }
     
+    g_free (uri);
     g_list_free (keys);
 }
 
@@ -397,10 +490,20 @@ sign_activate (GtkWidget *widget, SeahorseWidget *swidget)
 {
     SeahorseKey *skey;
     guint uid;
+    GList *keys = NULL;
 
     skey = get_selected_key (swidget, &uid);
-    if (SEAHORSE_IS_PGP_KEY (skey))
+    if (SEAHORSE_IS_PGP_KEY (skey)) {
         seahorse_sign_show (SEAHORSE_PGP_KEY (skey), uid);
+        
+#ifdef WITH_KEYSERVER
+        if (seahorse_gconf_get_boolean(AUTOSYNC_KEY) == TRUE) {
+            keys = g_list_append (keys, skey);
+            seahorse_keyserver_sync (keys);
+            g_list_free(keys);
+        }
+#endif
+    }
 }
 
 static void
@@ -428,7 +531,7 @@ sync_activate (GtkWidget *widget, SeahorseWidget *swidget)
     }
     
     if (keys == NULL)
-        keys = seahorse_context_find_keys (SCTX_APP (), SKEY_PGP, SKEY_LOC_LOCAL, 0);
+        keys = seahorse_context_find_keys (SCTX_APP (), SKEY_PGP, 0, SKEY_LOC_LOCAL);
     seahorse_keyserver_sync_show (keys);
     g_list_free (keys);
 #endif
@@ -464,7 +567,7 @@ delete_activate (GtkWidget *widget, SeahorseWidget *swidget)
     /* Special behavior for a single selection */
     if (g_list_length (keys) == 1) {
         SeahorseKey *skey;
-        guint uid;
+        guint uid = 0;
         
         skey = get_selected_key (swidget, &uid);
         if (uid > 0 && SEAHORSE_IS_PGP_KEY (skey)) 
@@ -504,10 +607,20 @@ collapse_all_activate (GtkMenuItem *item, SeahorseWidget *swidget)
 	    gtk_tree_view_collapse_all (view);
 }
 
+/* Makes URL in About Dialog Clickable */
+static void about_dialog_activate_link_cb (GtkAboutDialog *about,
+                                           const gchar *url,
+                                           gpointer data)
+{
+	gnome_url_show (url, NULL);
+}
+
 /* Shows about dialog */
 static void
 about_activate (GtkWidget *widget, SeahorseWidget *swidget)
 {
+    static gboolean been_here = FALSE;
+    
 	static const gchar *authors[] = {
 		"Jacob Perkins <jap1@users.sourceforge.net>",
 		"Jose Carlos Garcia Sogo <jsogo@users.sourceforge.net>",
@@ -533,6 +646,12 @@ about_activate (GtkWidget *widget, SeahorseWidget *swidget)
         NULL
     };
 
+    if (!been_here)
+	{
+		been_here = TRUE;
+		gtk_about_dialog_set_url_hook (about_dialog_activate_link_cb, NULL, NULL);
+	}
+	
 	gtk_show_about_dialog (NULL, 
                 "name", _("seahorse"),
                 "version", VERSION,
@@ -655,34 +774,22 @@ static void
 target_drag_data_received (GtkWidget *widget, GdkDragContext *context, gint x, gint y,
                            GtkSelectionData *data, guint info, guint time, SeahorseWidget *swidget)
 {
-    SeahorseKeySource *sksrc;
-    gint keys = 0;
-    GError *err = NULL;
     gchar** uris;
     gchar** u;
     
     DBG_PRINT(("DragDataReceived -->\n"));
     g_return_if_fail (data != NULL);
     
-    sksrc = seahorse_context_find_key_source (SCTX_APP (), SKEY_PGP, SKEY_LOC_LOCAL);
-    g_return_if_fail (sksrc && SEAHORSE_IS_PGP_SOURCE (sksrc));
-    
     switch(info) {
     case TEXT_PLAIN:
-        keys = seahorse_op_import_text (SEAHORSE_PGP_SOURCE (sksrc), 
-                                        (const gchar*)(data->data), &err);
+        import_text (swidget, (const gchar*)(data->data));
         break;
         
     case TEXT_URIS:
         uris = g_strsplit ((const gchar*)data->data, "\n", 0);
-        for(u = uris; *u; u++) {
+        for(u = uris; *u; u++)
             g_strstrip (*u);
-            if ((*u)[0]) { /* Make sure it's not an empty line */
-                keys += seahorse_op_import_file (SEAHORSE_PGP_SOURCE (sksrc), *u, &err);  
-                if (err != NULL)
-                    break;
-            }
-        }
+        import_files (swidget, (const gchar**)uris);
         g_strfreev (uris);
         break;
         
@@ -691,13 +798,6 @@ target_drag_data_received (GtkWidget *widget, GdkDragContext *context, gint x, g
         break;
     } 
 
-    if (err != NULL)
-        seahorse_util_handle_error (err, _("Couldn't import key from \"%s\""),
-                                    seahorse_util_uri_get_last (*u));
-    else
-        set_numbered_status (swidget, _("Imported %d key"), 
-                                      _("Imported %d keys"), keys);
-    
     DBG_PRINT(("DragDataReceived <--\n"));
 }
 
@@ -757,7 +857,7 @@ help_activate (GtkWidget *widget, SeahorseWidget *swidget)
 
 /* BUILDING THE MAIN WINDOW ------------------------------------------------- */
 
-static GtkActionEntry ui_entries[] = {
+static const GtkActionEntry ui_entries[] = {
     
     /* Top menu items */
     { "key-menu", NULL, N_("_Key") },
@@ -793,7 +893,7 @@ static GtkActionEntry ui_entries[] = {
             N_("Show Seahorse help"), G_CALLBACK (help_activate) }, 
 };
 
-static GtkActionEntry key_entries[] = {
+static const GtkActionEntry key_entries[] = {
     { "key-properties", GTK_STOCK_PROPERTIES, N_("P_roperties"), NULL,
             N_("Show key properties"), G_CALLBACK (properties_activate) }, 
     { "key-export-file", GTK_STOCK_SAVE_AS, N_("E_xport..."), NULL,
@@ -804,17 +904,17 @@ static GtkActionEntry key_entries[] = {
             N_("Delete selected keys"), G_CALLBACK (delete_activate) }, 
 };
 
-static GtkActionEntry pgp_entries[] = {
+static const GtkActionEntry pgp_entries[] = {
     { "key-sign", GTK_STOCK_INDEX, N_("_Sign..."), NULL,
             N_("Sign public key"), G_CALLBACK (sign_activate) }, 
 };
 
-static GtkActionEntry ssh_entries[] = {
+static const GtkActionEntry ssh_entries[] = {
     { "remote-upload-ssh", NULL, N_("Configure SSH Key for Computer..."), "",
             N_("Send public SSH key to another machine, and enable logins using that key."), G_CALLBACK (setup_sshkey_activate) },
 };
 
-static GtkActionEntry keyserver_entries[] = {
+static const GtkActionEntry keyserver_entries[] = {
     { "remote-find", GTK_STOCK_FIND, N_("_Find Remote Keys..."), "",
             N_("Search for keys on a key server"), G_CALLBACK (search_activate) }, 
     { "remote-sync", GTK_STOCK_REFRESH, N_("_Sync and Publish Keys..."), "",
@@ -858,22 +958,24 @@ GtkWindow*
 seahorse_key_manager_show (SeahorseOperation *op)
 {
     GtkWindow *win;
-	SeahorseWidget *swidget;
+    SeahorseWidget *swidget;
     GtkWidget *w;
     GtkActionGroup *actions;
     GtkAction *action;
-	
-	swidget = seahorse_widget_new ("key-manager");
+    
+    swidget = seahorse_widget_new ("key-manager");
     win = GTK_WINDOW (glade_xml_get_widget (swidget->xml, "key-manager"));
     
     /* General normal actions */
     actions = gtk_action_group_new ("main");
+    gtk_action_group_set_translation_domain (actions, NULL);
     gtk_action_group_add_actions (actions, ui_entries, 
                                   G_N_ELEMENTS (ui_entries), swidget);
     seahorse_widget_add_actions (swidget, actions);
     
     /* Actions that are allowed on all keys */
     actions = gtk_action_group_new ("key");
+    gtk_action_group_set_translation_domain (actions, NULL);
     gtk_action_group_add_actions (actions, key_entries, 
                                   G_N_ELEMENTS (key_entries), swidget);
     seahorse_widget_add_actions (swidget, actions);
@@ -885,30 +987,33 @@ seahorse_key_manager_show (SeahorseOperation *op)
 
     /* Actions for pgp keys */
     actions = gtk_action_group_new ("pgp");
+    gtk_action_group_set_translation_domain (actions, NULL);
     gtk_action_group_add_actions (actions, pgp_entries,
                                   G_N_ELEMENTS (pgp_entries), swidget);
     seahorse_widget_add_actions (swidget, actions);                              
     
     /* Actions for SSH keys */
     actions = gtk_action_group_new ("ssh");
+    gtk_action_group_set_translation_domain (actions, NULL);
     gtk_action_group_add_actions (actions, ssh_entries,
                                   G_N_ELEMENTS (ssh_entries), swidget);
     seahorse_widget_add_actions (swidget, actions);
-	
+    
 #ifndef WITH_SSH
     gtk_action_group_set_visible (actions, FALSE);
 #endif
 
     /* Actions for keyservers */
     actions = gtk_action_group_new ("keyserver");
+    gtk_action_group_set_translation_domain (actions, NULL);
     gtk_action_group_add_actions (actions, keyserver_entries, 
                                   G_N_ELEMENTS (keyserver_entries), swidget);
     seahorse_widget_add_actions (swidget, actions);                                  
-	
+    
 #ifndef WITH_KEYSERVER
-	gtk_action_group_set_visible (actions, FALSE);
+    gtk_action_group_set_visible (actions, FALSE);
 #endif
- 	
+    
 	/* close event */
 	glade_xml_signal_connect_data (swidget->xml, "quit_event",
 		G_CALLBACK (delete_event), swidget);
