@@ -29,13 +29,19 @@
 #include "seahorse-gpgmex.h"
 #include "seahorse-key-dialogs.h"
 #include "seahorse-key-widget.h"
-#include "seahorse-op.h"
 #include "seahorse-util.h"
 #include "seahorse-key.h"
 #include "seahorse-pgp-key.h"
 #include "seahorse-pgp-key-op.h"
 #include "seahorse-gtkstock.h"
 #include "seahorse-windows.h"
+#include "seahorse-vfs-data.h"
+#include "seahorse-key-model.h"
+#include "seahorse-gconf.h"
+
+#ifdef WITH_KEYSERVER
+#include "seahorse-keyserver-sync.h"
+#endif
 
 #define NOTEBOOK "notebook"
 
@@ -95,6 +101,44 @@ get_selected_row (SeahorseWidget *swidget, const gchar *gladeid, guint column)
     return index;
 }
 
+static void
+signature_row_activated(GtkTreeView *treeview, GtkTreePath *path, GtkTreeViewColumn *arg2, 
+                        SeahorseWidget *swidget)
+{
+    SeahorseKey *skey = NULL;
+    GtkTreeModel *model;
+    GtkTreeIter iter;
+    
+    model = gtk_tree_view_get_model (treeview);
+    
+    if (GTK_IS_TREE_MODEL_FILTER (model)) 
+        model = gtk_tree_model_filter_get_model (GTK_TREE_MODEL_FILTER (model));
+        
+    g_return_if_fail (gtk_tree_model_get_iter (model, &iter, path));
+    
+    skey = seahorse_key_model_get_row_key (SEAHORSE_KEY_MODEL (model), &iter);
+    
+	if (skey != NULL) {
+	   if (SEAHORSE_IS_PGP_KEY (skey))
+	       seahorse_key_properties_new (SEAHORSE_PGP_KEY (skey));
+    }
+}
+
+static GSList*
+unique_slist_strings (GSList *keyids)
+{
+    GSList *l;
+    
+    keyids = g_slist_sort (keyids, (GCompareFunc)g_ascii_strcasecmp);
+    for (l = keyids; l; l = g_slist_next (l)) {
+        while (l->next && l->data && l->next->data && 
+               g_ascii_strcasecmp (l->data, l->next->data) == 0)
+            keyids = g_slist_delete_link (keyids, l->next);
+    }    
+    
+    return keyids;
+}
+
 /* -----------------------------------------------------------------------------
  * NAMES PAGE (PRIVATE KEYS)
  */
@@ -103,8 +147,7 @@ enum {
     UIDSIG_INDEX,
     UIDSIG_ICON,
     UIDSIG_NAME,
-    UIDSIG_EMAIL,
-    UIDSIG_COMMENT,
+    UIDSIG_KEYID,
     UIDSIG_N_COLUMNS
 };
 
@@ -112,8 +155,7 @@ const GType uidsig_columns[] = {
     G_TYPE_INT,     /* index */
     G_TYPE_STRING,  /* icon */
     G_TYPE_STRING,  /* name */
-    G_TYPE_STRING,  /* email */
-    G_TYPE_STRING   /* comment */
+    G_TYPE_STRING   /* keyid */
 };
 
 static gint
@@ -163,18 +205,46 @@ names_sign_clicked (GtkWidget *widget, SeahorseWidget *swidget)
 {
     SeahorseKey *skey;
     gint index;
+    GList *keys = NULL;
 
     skey = SEAHORSE_KEY_WIDGET (swidget)->skey;
     index = names_get_selected_uid (swidget);
 
-    if (index >= 1) 
+    if (index >= 1) {
         seahorse_sign_show (SEAHORSE_PGP_KEY (skey), index - 1);
+
+#ifdef WITH_KEYSERVER
+        if (seahorse_gconf_get_boolean(AUTOSYNC_KEY) == TRUE) {
+            keys = g_list_append (keys, skey);
+            seahorse_keyserver_sync (keys);
+            g_list_free(keys);
+        }
+#endif
+    }
 }
 
 static void 
 names_revoke_clicked (GtkWidget *widget, SeahorseWidget *swidget)
 {
     /* TODO: */
+/*    SeahorseKey *skey;
+    gint index;
+    Glist *keys = NULL;
+
+    skey = SEAHORSE_KEY_WIDGET (swidget)->skey;
+    index = names_get_selected_uid (swidget);
+
+    if (index >= 1) {
+        seahorse_revoke_show (SEAHORSE_PGP_KEY (skey), index - 1);
+
+#ifdef WITH_KEYSERVER
+        if (seahorse_gconf_get_boolean(AUTOSYNC_KEY) == TRUE) {
+            keys = g_list_append (keys, skey);
+            seahorse_keyserver_sync (keys);
+            g_list_free(keys);
+        }
+#endif
+    }*/
 }
 
 static void
@@ -188,20 +258,96 @@ update_names (GtkTreeSelection *selection, SeahorseWidget *swidget)
     show_glade_widget (swidget, "names-revoke-button", FALSE);
 }
 
+/* Is called whenever a signature key changes, to update row */
+static void
+names_update_row (SeahorseKeyModel *skmodel, SeahorseKey *skey, 
+                  GtkTreeIter *iter, SeahorseWidget *swidget)
+{
+    const gchar *icon;
+    gchar *name;
+    
+    /* Always use the most preferred key for this keyid */
+    if (skey->preferred) {
+        while (skey->preferred)
+            skey = skey->preferred;
+        seahorse_key_model_set_row_key (skmodel, iter, skey);
+    }
+
+    icon = seahorse_key_get_location (skey) < SKEY_LOC_LOCAL ? 
+                GTK_STOCK_DIALOG_QUESTION : SEAHORSE_STOCK_SIGN;
+    name = seahorse_key_get_display_name (skey);
+    
+    gtk_tree_store_set (GTK_TREE_STORE (skmodel), iter,
+                        UIDSIG_INDEX, -1,
+                        UIDSIG_ICON, icon,
+                        UIDSIG_NAME, name ? name : _("[Unknown]"),
+                        UIDSIG_KEYID, seahorse_key_get_short_keyid (skey), -1);
+            
+    g_free (name);
+}
+
+static void
+names_populate (SeahorseWidget *swidget, GtkTreeStore *store, SeahorsePGPKey *pkey)
+{
+    SeahorseKey *skey;
+    gpgme_user_id_t uid;
+    gpgme_key_sig_t sig;
+    GtkTreeIter uiditer, sigiter;
+    gchar *name;
+    const gchar *keyid;
+    GSList *rawids = NULL;
+    int i, j;
+    GList *keys, *l;
+
+    keyid = seahorse_pgp_key_get_id (pkey->pubkey, 0);
+    
+    /* Insert all the fun-ness */
+    for (i = 1, uid = pkey->pubkey->uids; uid; uid = uid->next, i++) {
+
+        name = seahorse_pgp_key_get_userid (pkey, i - 1);
+
+        gtk_tree_store_append (store, &uiditer, NULL);
+        gtk_tree_store_set (store, &uiditer,  
+                            UIDSIG_INDEX, i,
+                            UIDSIG_ICON, SEAHORSE_STOCK_PERSON,
+                            UIDSIG_NAME, name,
+                            -1);
+        
+        g_free (name);
+        
+        
+        /* Build a list of all the keyids */
+        for (j = 1, sig = uid->signatures; sig; sig = sig->next, j++) {
+            /* Never show self signatures, they're implied */
+            if (strcmp (sig->keyid, keyid) == 0)
+                continue;
+            rawids = g_slist_prepend (rawids, sig->keyid);
+        }
+        
+        /* Pass it to 'DiscoverKeys' for resolution/download */
+        keys = seahorse_context_discover_keys (SCTX_APP (), SKEY_PGP, rawids);
+        g_slist_free (rawids);
+        rawids = NULL;
+        
+        /* Add the keys to the store */
+        for (l = keys; l; l = g_list_next (l)) {
+            skey = SEAHORSE_KEY (l->data);
+            gtk_tree_store_append (store, &sigiter, &uiditer);
+            
+            /* This calls the 'update-row' callback, to set the values for the key */
+            seahorse_key_model_set_row_key (SEAHORSE_KEY_MODEL (store), &sigiter, skey);
+        }
+    } 
+}
+
 static void
 do_names (SeahorseWidget *swidget)
 {
     SeahorseKey *skey;
     SeahorsePGPKey *pkey;
-    gpgme_user_id_t uid;
-    gpgme_key_sig_t sig;
     GtkWidget *widget;
     GtkTreeStore *store;
     GtkCellRenderer *renderer;
-    GtkTreeIter uiditer, sigiter;
-    gchar *name, *email, *comment;
-    const gchar *keyid;
-    int i, j;
     
     skey = SEAHORSE_KEY_WIDGET (swidget)->skey;
     pkey = SEAHORSE_PGP_KEY (skey);
@@ -220,7 +366,8 @@ do_names (SeahorseWidget *swidget)
     } else {
         
         /* This is our first time so create a store */
-        store = gtk_tree_store_newv (UIDSIG_N_COLUMNS, (GType*)uidsig_columns);
+        store = GTK_TREE_STORE (seahorse_key_model_new (UIDSIG_N_COLUMNS, (GType*)uidsig_columns));
+        g_signal_connect (store, "update-row", G_CALLBACK (names_update_row), swidget);
 
         /* Make the columns for the view */
         renderer = gtk_cell_renderer_pixbuf_new ();
@@ -230,60 +377,15 @@ do_names (SeahorseWidget *swidget)
                                                      "stock-id", UIDSIG_ICON, NULL);
 
         gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widget), 
-                                                     -1, _("Name"), gtk_cell_renderer_text_new (), 
+                                                     -1, _("Name/Email"), gtk_cell_renderer_text_new (), 
                                                      "text", UIDSIG_NAME, NULL);
 
-        gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widget),
-                                                     -1, _("Email"), gtk_cell_renderer_text_new (), 
-                                                     "text", UIDSIG_EMAIL, NULL);
-
-        gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widget),
-                                                     -1, _("Comment"), gtk_cell_renderer_text_new (), 
-                                                     "text", UIDSIG_COMMENT, NULL);
+        gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widget), 
+                                                     -1, _("Signature ID"), gtk_cell_renderer_text_new (), 
+                                                     "text", UIDSIG_KEYID, NULL);
     }
 
-    keyid = seahorse_pgp_key_get_id (pkey->pubkey, 0);
-    
-    /* Insert all the fun-ness */
-    for (i = 1, uid = pkey->pubkey->uids; uid; uid = uid->next, i++) {
-
-        name = seahorse_pgp_key_get_userid_name (pkey, i - 1);
-        email = seahorse_pgp_key_get_userid_email (pkey, i - 1);
-        comment = seahorse_pgp_key_get_userid_comment (pkey, i - 1);
-
-        gtk_tree_store_append (store, &uiditer, NULL);
-        gtk_tree_store_set (store, &uiditer,  
-                            UIDSIG_INDEX, i,
-                            UIDSIG_ICON, SEAHORSE_STOCK_PERSON,
-                            UIDSIG_NAME, name,
-                            UIDSIG_EMAIL, email,
-                            UIDSIG_COMMENT, comment,
-                            -1);
-        
-        g_free (name);
-        g_free (email);
-        g_free (comment);
-        
-        for (j = 1, sig = uid->signatures; sig; sig = sig->next, j++) {
-            
-            /* Never show self signatures, they're implied */
-            if (strcmp (sig->keyid, keyid) == 0)
-                continue;
-            
-            seahorse_pgp_key_get_signature_text (pkey, sig, &name, &email, NULL);
-            
-            gtk_tree_store_append (store, &sigiter, &uiditer);
-            gtk_tree_store_set (store, &sigiter,
-                                UIDSIG_INDEX, -1,
-                                UIDSIG_ICON, SEAHORSE_STOCK_SIGN,
-                                UIDSIG_NAME, name ? name : _("[Unknown]"),
-                                UIDSIG_EMAIL, email,
-                                UIDSIG_COMMENT, sig->keyid, -1);
-            
-            g_free (name);
-            g_free (email);
-        }
-    } 
+    names_populate (swidget, store, pkey);
     
     gtk_tree_view_set_model (GTK_TREE_VIEW (widget), GTK_TREE_MODEL(store));
     gtk_tree_view_expand_all (GTK_TREE_VIEW (widget));
@@ -313,6 +415,8 @@ do_names_signals (SeahorseWidget *swidget)
             G_CALLBACK (names_delete_clicked), swidget);
     glade_xml_signal_connect_data (swidget->xml, "on_names_revoke",
             G_CALLBACK (names_revoke_clicked), swidget);
+    glade_xml_signal_connect_data (swidget->xml, "on_names_tree_row_activated",
+                                   G_CALLBACK (signature_row_activated), swidget);
     
     widget = seahorse_widget_get_widget (swidget, "names-tree");
     g_return_if_fail (widget != NULL);
@@ -324,6 +428,60 @@ do_names_signals (SeahorseWidget *swidget)
  * PHOTO ID AREA
  */
 
+/* drag-n-drop uri data */
+enum {
+    TARGET_URI, 
+};
+
+static GtkTargetEntry target_list[] = {
+    { "text/uri-list", 0, TARGET_URI } };
+
+static guint n_targets = G_N_ELEMENTS (target_list);
+
+static void
+owner_photo_drag_received (GtkWidget *widget, GdkDragContext *context, gint x, gint y,
+                           GtkSelectionData *sel_data, guint target_type, guint time,
+                           SeahorseWidget *swidget)
+{       
+    gboolean dnd_success = FALSE;
+    SeahorsePGPKey *pkey;
+    gchar **uri_list;
+    gint len = 0;
+    gchar *uri;
+    
+    pkey = SEAHORSE_PGP_KEY (SEAHORSE_KEY_WIDGET (swidget)->skey);
+    g_assert (SEAHORSE_IS_PGP_KEY (pkey)); 
+    
+    /* 
+     * This needs to be improved, support should be added for remote images 
+     * and there has to be a better way to get rid of the trailing \r\n appended
+     * to the end of the path after the call to g_filename_from_uri
+     */
+    if((sel_data != NULL) && (sel_data->length >= 0)) {
+        g_return_if_fail (target_type == TARGET_URI);
+        
+        uri_list = gtk_selection_data_get_uris (sel_data);
+        while (uri_list && uri_list[len]) {
+                
+            uri = g_filename_from_uri (uri_list[len], NULL, NULL);
+            if (!uri)
+                continue;
+                
+            dnd_success = seahorse_photo_add (pkey, GTK_WINDOW (seahorse_widget_get_top (swidget)),
+                                              uri);
+            g_free (uri);
+            
+            if (!dnd_success)
+                break;
+            len++;
+        }
+        
+        g_strfreev (uri_list);
+    }
+    
+    gtk_drag_finish (context, dnd_success, FALSE, time);
+}
+
 static void
 owner_photo_add_button_clicked (GtkWidget *widget, SeahorseWidget *swidget)
 {
@@ -332,7 +490,7 @@ owner_photo_add_button_clicked (GtkWidget *widget, SeahorseWidget *swidget)
     pkey = SEAHORSE_PGP_KEY (SEAHORSE_KEY_WIDGET (swidget)->skey);
     g_assert (SEAHORSE_IS_PGP_KEY (pkey));
     
-    if (seahorse_photo_add (pkey, GTK_WINDOW (seahorse_widget_get_top (swidget))))
+    if (seahorse_photo_add (pkey, GTK_WINDOW (seahorse_widget_get_top (swidget)), NULL))
         g_object_set_data (G_OBJECT (swidget), "current-photoid", NULL);
 }
  
@@ -553,6 +711,7 @@ do_owner_signals (SeahorseWidget *swidget)
 { 
     SeahorseKey *skey;
     SeahorseKeyEType etype;
+    GtkWidget *frame;
     
     skey = SEAHORSE_KEY_WIDGET (swidget)->skey;
     etype = seahorse_key_get_etype (skey);
@@ -575,6 +734,11 @@ do_owner_signals (SeahorseWidget *swidget)
                 G_CALLBACK (owner_photo_delete_button_clicked), swidget);
         glade_xml_signal_connect_data (swidget->xml, "on_passphrase_button",
                 G_CALLBACK (owner_passphrase_button_clicked), swidget);
+        glade_xml_signal_connect_data (swidget->xml, "on_owner_photo_drag_received",
+                G_CALLBACK (owner_photo_drag_received), swidget);
+        frame = glade_xml_get_widget (swidget->xml, "owner-photo-frame");
+        gtk_drag_dest_set (frame, GTK_DEST_DEFAULT_MOTION | GTK_DEST_DEFAULT_HIGHLIGHT |
+                           GTK_DEST_DEFAULT_DROP, target_list, n_targets, GDK_ACTION_COPY);
     } else {
         show_glade_widget (swidget, "owner-photo-add-button", FALSE);
         show_glade_widget (swidget, "owner-photo-delete-button", FALSE);
@@ -801,31 +965,48 @@ trust_changed (GtkComboBox *selection, SeahorseWidget *swidget)
 static void
 details_export_button_clicked (GtkWidget *widget, SeahorseWidget *swidget)
 {
-	SeahorseKey *skey;
-	GtkWidget *dialog;
-	gchar* uri = NULL;
-	GError *err = NULL;
-	GList *keys = NULL;
-	
-	skey = SEAHORSE_KEY_WIDGET (swidget)->skey;
-	keys = g_list_prepend (keys, skey);
-	
-	dialog = seahorse_util_chooser_save_new (_("Export Complete Key"), 
-	                                         GTK_WINDOW (seahorse_widget_get_top (swidget)));
-	seahorse_util_chooser_show_key_files (dialog);
-	seahorse_util_chooser_set_filename (dialog, keys);
-	
-	uri = seahorse_util_chooser_save_prompt (dialog);
-	if(uri) {
-		seahorse_op_export_file (keys, TRUE, uri, &err);
-
-		if (err != NULL) {
-			seahorse_util_handle_error (err, _("Couldn't export key to \"%s\""),
-										seahorse_util_uri_get_last (uri));
-		}
-		g_free (uri);
-	}
-	g_list_free (keys);
+    SeahorseKeySource *sksrc;
+    SeahorseOperation *op;
+    SeahorseKey *skey;
+    GtkWidget *dialog;
+    gchar* uri = NULL;
+    GError *err = NULL;
+    gpgme_data_t data;
+    GList *keys = NULL;
+    
+    skey = SEAHORSE_KEY_WIDGET (swidget)->skey;
+    keys = g_list_prepend (NULL, skey);
+    
+    dialog = seahorse_util_chooser_save_new (_("Export Complete Key"), 
+                                             GTK_WINDOW (seahorse_widget_get_top (swidget)));
+    seahorse_util_chooser_show_key_files (dialog);
+    seahorse_util_chooser_set_filename (dialog, keys);
+    
+    uri = seahorse_util_chooser_save_prompt (dialog);
+    if(!uri) 
+        return;
+    
+    sksrc = seahorse_key_get_source (skey);
+    g_assert (SEAHORSE_IS_KEY_SOURCE (sksrc));
+    
+    data = seahorse_vfs_data_create (uri, SEAHORSE_VFS_WRITE, &err);
+    
+    if (data) {
+        op = seahorse_key_source_export (sksrc, keys, TRUE, data);
+    
+        seahorse_operation_wait (op);
+        gpgmex_data_release (data);
+        
+        if (!seahorse_operation_is_successful (op))
+            seahorse_operation_steal_error (op, &err);
+    }
+    
+    if (err)
+        seahorse_util_handle_error (err, _("Couldn't export key to \"%s\""),
+                                    seahorse_util_uri_get_last (uri));
+    
+    g_list_free (keys);
+    g_free (uri);
 }
 
 /*
@@ -1043,10 +1224,10 @@ do_details (SeahorseWidget *swidget)
  */
 
 enum {
-    SIGN_KEY_ID,
+    SIGN_ICON,
     SIGN_NAME,
-    SIGN_EMAIL,
-    SIGN_COMMENT,
+    SIGN_KEYID,
+    SIGN_TRUSTED,
     SIGN_N_COLUMNS
 };
 
@@ -1054,7 +1235,7 @@ const GType sign_columns[] = {
     G_TYPE_STRING,
     G_TYPE_STRING,
     G_TYPE_STRING,
-    G_TYPE_STRING
+    G_TYPE_BOOLEAN
 };
 
 static void
@@ -1107,93 +1288,107 @@ trust_complete_toggled (GtkToggleButton *toggle, SeahorseWidget *swidget)
     }
 }
 
+/* Is called whenever a signature key changes */
 static void
-signatures_populate_model (SeahorseWidget *swidget, GtkListStore *store)
+trust_update_row (SeahorseKeyModel *skmodel, SeahorseKey *skey, 
+                  GtkTreeIter *iter, SeahorseWidget *swidget)
+{
+    gboolean trusted = FALSE;
+    const gchar *icon;
+    gchar *name;    
+    
+    /* Always use the most preferred key for this keyid */
+    if (skey->preferred) {
+        while (skey->preferred)
+            skey = skey->preferred;
+        seahorse_key_model_set_row_key (skmodel, iter, skey);
+    }
+
+    if (seahorse_key_get_etype (skey) == SKEY_PRIVATE) 
+        trusted = TRUE;
+    else if (seahorse_key_get_flags (skey) & SKEY_FLAG_TRUSTED)
+        trusted = TRUE;
+    
+    icon = seahorse_key_get_location (skey) < SKEY_LOC_LOCAL ? 
+                GTK_STOCK_DIALOG_QUESTION : SEAHORSE_STOCK_SIGN;
+    name = seahorse_key_get_display_name (skey);
+    
+    gtk_tree_store_set (GTK_TREE_STORE (skmodel), iter,
+                        SIGN_ICON, icon,
+                        SIGN_NAME, name ? name : _("[Unknown]"),
+                        SIGN_KEYID, seahorse_key_get_short_keyid (skey),
+                        SIGN_TRUSTED, trusted,
+                        -1);
+            
+    g_free (name);
+}
+
+static void
+signatures_populate_model (SeahorseWidget *swidget, SeahorseKeyModel *skmodel)
 {
     SeahorseKey *skey;
     SeahorsePGPKey *pkey;
     GtkTreeIter iter;
     GtkWidget *widget;
-    gboolean trusted_only = TRUE;
-    guint sigtype;
-    const gchar *name = NULL;
     const gchar *keyid;
     gpgme_user_id_t uid;
     gpgme_key_sig_t sig;
     gboolean have_sigs = FALSE;
+    GSList *rawids = NULL;
+    GList *keys, *l;
 
-    skey = SEAHORSE_KEY_WIDGET (swidget)->skey;
-    pkey = SEAHORSE_PGP_KEY (skey);
-
-    widget = glade_xml_get_widget (swidget->xml, "signatures-toggle");
-    if (widget)
-        trusted_only = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
-    
+    pkey = SEAHORSE_PGP_KEY (SEAHORSE_KEY_WIDGET (swidget)->skey);
     widget = glade_xml_get_widget (swidget->xml, "signatures-tree");
     
-    if (store) {
-        keyid = seahorse_pgp_key_get_id (pkey->pubkey, 0);
+    keyid = seahorse_pgp_key_get_id (pkey->pubkey, 0);
 
-        for (uid = pkey->pubkey->uids; uid; uid = uid->next) {
-            for (sig = uid->signatures; sig; sig = sig->next) {
-                
-                /* Never show self signatures, they're implied */
-                if (strcmp (sig->keyid, keyid) == 0)
-                    continue;
-                
-                have_sigs = TRUE;
-                
-                /* Find any trusted signatures */
-                sigtype = seahorse_pgp_key_get_sigtype (pkey, sig);
-                if (trusted_only && !(sigtype & SKEY_PGPSIG_TRUSTED || sigtype & SKEY_PGPSIG_PERSONAL))
-                    continue;
-
-                gtk_list_store_append (store, &iter);
-                
-                if (!name)
-                    gtk_list_store_set (store, &iter, 
-                                        SIGN_KEY_ID, sig->keyid,
-                                        SIGN_NAME, sig->name, 
-                                        SIGN_EMAIL, sig->email, 
-                                        SIGN_COMMENT, sig->comment, 
-                                        -1);
-                
-                else if (g_str_equal (name, sig->name))
-                    gtk_list_store_set (store, &iter, 
-                                        SIGN_KEY_ID, sig->keyid,
-                                        SIGN_NAME, sig->name, 
-                                        SIGN_EMAIL, sig->email, 
-                                        SIGN_COMMENT, sig->comment, 
-                                        -1);
-            }
+    /* Build a list of all the keyids */        
+    for (uid = pkey->pubkey->uids; uid; uid = uid->next) {
+        for (sig = uid->signatures; sig; sig = sig->next) {
+            /* Never show self signatures, they're implied */
+            if (strcmp (sig->keyid, keyid) == 0)
+                continue;
+            have_sigs = TRUE;
+            rawids = g_slist_prepend (rawids, sig->keyid);
         }
     }
-
-    gtk_tree_view_set_model (GTK_TREE_VIEW (widget), GTK_TREE_MODEL(store));
     
     /* Only show signatures area when there are signatures */
     seahorse_widget_set_visible (swidget, "signatures-area", have_sigs);
+
+    if (skmodel) {
+    
+        /* String out duplicates */
+        rawids = unique_slist_strings (rawids);
+        
+        /* Pass it to 'DiscoverKeys' for resolution/download */
+        keys = seahorse_context_discover_keys (SCTX_APP (), SKEY_PGP, rawids);
+        g_slist_free (rawids);
+        rawids = NULL;
+        
+        /* Add the keys to the store */
+        for (l = keys; l; l = g_list_next (l)) {
+            skey = SEAHORSE_KEY (l->data);
+            gtk_tree_store_append (GTK_TREE_STORE (skmodel), &iter, NULL);
+            
+            /* This calls the 'update-row' callback, to set the values for the key */
+            seahorse_key_model_set_row_key (SEAHORSE_KEY_MODEL (skmodel), &iter, skey);
+        }
+    }
 }
 
-/* 
- * The signatures toggle filters the view for the user so that only signatures 
- * in the collected keys appear. This way the user won't have to filter 
- * through meaningless KeyIDs. Self signatures on personal keys should be 
- * assumed, don't display those by default either.
- */
+/* Refilter when the user toggles the 'only show trusted' checkbox */
 static void
-trusted_toggled (GtkToggleButton *toggle, SeahorseWidget *swidget)
+trusted_toggled (GtkToggleButton *toggle, GtkTreeModelFilter *filter)
 {
-    GtkWidget *widget;
-    GtkListStore *store;
-
-    widget = glade_xml_get_widget (swidget->xml, "signatures-tree");
-    store = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (widget)));
-
-    gtk_list_store_clear (store);
-    signatures_populate_model (swidget, store);	
+    /* Set flag on the store */
+    GtkTreeModel *model = gtk_tree_model_filter_get_model (filter);
+    g_object_set_data (G_OBJECT (model), "only-trusted", 
+                GINT_TO_POINTER (gtk_toggle_button_get_active (toggle)));
+    gtk_tree_model_filter_refilter (filter);
 }
 
+/* Add a signature */
 static void 
 trust_sign_clicked (GtkWidget *widget, SeahorseWidget *swidget)
 {
@@ -1206,7 +1401,6 @@ do_trust_signals (SeahorseWidget *swidget)
 {
     SeahorseKey *skey;
     SeahorseKeyEType etype;
-    GtkWidget *widget;
 
     skey = SEAHORSE_KEY_WIDGET (swidget)->skey;
     etype = seahorse_key_get_etype (skey);
@@ -1232,16 +1426,23 @@ do_trust_signals (SeahorseWidget *swidget)
                                        G_CALLBACK (signatures_delete_button_clicked), swidget);
     }
 
-    glade_xml_signal_connect_data (swidget->xml, "on_signatures_toggle_toggled",
-                                   G_CALLBACK (trusted_toggled), swidget);
-    widget = glade_xml_get_widget (swidget->xml, "signatures-toggle");
-    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-    
     glade_xml_signal_connect_data (swidget->xml, "trust_marginal_toggled",
                                    G_CALLBACK (trust_marginal_toggled), swidget);
     glade_xml_signal_connect_data (swidget->xml, "trust_complete_toggled",
                                    G_CALLBACK (trust_complete_toggled), swidget);
-    
+                                   
+    glade_xml_signal_connect_data (swidget->xml, "on_signatures_tree_row_activated",
+                                   G_CALLBACK (signature_row_activated), swidget);
+}
+
+/* When the 'only display trusted' check is checked, hide untrusted rows */
+static gboolean
+trust_filter (GtkTreeModel *model, GtkTreeIter *iter, gpointer userdata)
+{
+    /* Read flag on the store */
+    gboolean trusted = FALSE;
+    gtk_tree_model_get (model, iter, SIGN_TRUSTED, &trusted, -1);
+    return !g_object_get_data (G_OBJECT (model), "only-trusted") || trusted;
 }
 
 static void 
@@ -1250,8 +1451,10 @@ do_trust (SeahorseWidget *swidget)
     SeahorseKey *skey;
     SeahorsePGPKey *pkey;
     GtkWidget *widget;
-    GtkListStore *store;
+    GtkTreeStore *store;
+    GtkTreeModelFilter *filter;
     gboolean sigpersonal;
+    GtkCellRenderer *renderer;
     guint keyloc;
     
     skey = SEAHORSE_KEY_WIDGET (swidget)->skey;
@@ -1262,7 +1465,7 @@ do_trust (SeahorseWidget *swidget)
         return;
     
     /* Remote keys */
-    if (keyloc != SKEY_LOC_LOCAL) {
+    if (keyloc < SKEY_LOC_LOCAL) {
         
         show_glade_widget (swidget, "manual-trust-area", FALSE);
         show_glade_widget (swidget, "manage-trust-area", TRUE);
@@ -1320,6 +1523,7 @@ do_trust (SeahorseWidget *swidget)
             break;
         
         default:
+            g_warning ("unknown trust value: %d", trust);
             g_assert_not_reached ();
             return;
         }
@@ -1359,30 +1563,47 @@ do_trust (SeahorseWidget *swidget)
     
     /* The actual signatures listing */
     widget = glade_xml_get_widget (swidget->xml, "signatures-tree");
-    store = GTK_LIST_STORE (gtk_tree_view_get_model (GTK_TREE_VIEW (widget)));
+    filter = GTK_TREE_MODEL_FILTER (gtk_tree_view_get_model (GTK_TREE_VIEW (widget)));
     
-    if (store) {
-        gtk_list_store_clear (store);
+    if (filter) {
+        store = GTK_TREE_STORE (gtk_tree_model_filter_get_model (filter));
+        gtk_tree_store_clear (store);
     
     /* First time create the store */
     } else {
-        store = gtk_list_store_newv (SIGN_N_COLUMNS, (GType*)sign_columns);
-
+        
+        /* Create a new SeahorseKeyModel store.... */
+        store = GTK_TREE_STORE (seahorse_key_model_new (SIGN_N_COLUMNS, (GType*)sign_columns));
+        g_signal_connect (store, "update-row", G_CALLBACK (trust_update_row), swidget);
+        
+        /* .... and a filter to go ontop of it */
+        filter = GTK_TREE_MODEL_FILTER (gtk_tree_model_filter_new (GTK_TREE_MODEL (store), NULL));
+        gtk_tree_model_filter_set_visible_func (filter, 
+                                (GtkTreeModelFilterVisibleFunc)trust_filter, NULL, NULL);
+        
+        /* Make the colunms for the view */
+        renderer = gtk_cell_renderer_pixbuf_new ();
+        g_object_set (renderer, "stock-size", GTK_ICON_SIZE_LARGE_TOOLBAR, NULL);
+        gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widget),
+                                                     -1, "", renderer,
+                                                     "stock-id", SIGN_ICON, NULL);
         gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widget), 
-                                                     -1, _("ID"), gtk_cell_renderer_text_new (), 
-                                                     "text", SIGN_KEY_ID, NULL);
-        gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widget), 
-                                                     -1, _("Name"), gtk_cell_renderer_text_new (), 
+                                                     -1, _("Name/Email"), gtk_cell_renderer_text_new (), 
                                                      "text", SIGN_NAME, NULL);
         gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widget), 
-                                                     -1, _("Email"), gtk_cell_renderer_text_new (), 
-                                                     "text", SIGN_EMAIL, NULL);
-        gtk_tree_view_insert_column_with_attributes (GTK_TREE_VIEW (widget), 
-                                                     -1, _("Comment"), gtk_cell_renderer_text_new (), 
-                                                     "text", SIGN_COMMENT, NULL);
+                                                     -1, _("Key ID"), gtk_cell_renderer_text_new (), 
+                                                     "text", SIGN_KEYID, NULL);
+        
+        gtk_tree_view_set_model (GTK_TREE_VIEW (widget), GTK_TREE_MODEL(filter));
+        
+        glade_xml_signal_connect_data (swidget->xml, "on_signatures_toggle_toggled",
+                                       G_CALLBACK (trusted_toggled), filter);
+        widget = glade_xml_get_widget (swidget->xml, "signatures-toggle");
+        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
+        trusted_toggled (GTK_TOGGLE_BUTTON (widget), filter);
     } 
 
-    signatures_populate_model (swidget, store);
+    signatures_populate_model (swidget, SEAHORSE_KEY_MODEL (store));
 }
 
 /* -----------------------------------------------------------------------------
@@ -1541,7 +1762,7 @@ seahorse_key_properties_new (SeahorsePGPKey *pkey)
     /* Don't trigger the import of remote keys if possible */
     if (!remote) {
         /* This causes the key source to get any specific info about the key */
-        seahorse_key_source_load_sync (sksrc, SKSRC_LOAD_KEY, seahorse_key_get_keyid (skey));
+        seahorse_key_source_load_sync (sksrc, SKSRC_LOAD_KEY, seahorse_key_get_keyid (skey), NULL);
         skey = seahorse_context_get_key (SCTX_APP(), sksrc, seahorse_key_get_keyid (skey));
         g_return_if_fail (skey != NULL);
     }
