@@ -24,7 +24,6 @@
 
 #include "seahorse-widget.h"
 #include "seahorse-util.h"
-#include "seahorse-op.h"
 #include "seahorse-gpgmex.h"
 #include "seahorse-context.h"
 #include "seahorse-windows.h"
@@ -32,6 +31,8 @@
 #include "seahorse-preferences.h"
 #include "seahorse-server-source.h"
 #include "seahorse-gconf.h"
+#include "seahorse-transfer-operation.h"
+#include "seahorse-keyserver-sync.h"
 
 static void 
 sync_import_complete (SeahorseOperation *op, SeahorseKeySource *sksrc)
@@ -49,25 +50,9 @@ static void
 sync_export_complete (SeahorseOperation *op, SeahorseKeySource *sksrc)
 {
     GError *err = NULL;
+    gchar *keyserver;
 
-    if (seahorse_operation_is_successful (op)) {
-        SeahorseKeySource *lsrc;
-        gpgme_data_t data;
-        
-        data = seahorse_operation_get_result (op);
-        g_return_if_fail (data != NULL);
-
-        /* Now import it into the local key source */
-        lsrc = seahorse_context_find_key_source (SCTX_APP (), SKEY_PGP, SKEY_LOC_LOCAL);
-        g_return_if_fail (lsrc);
-        
-        if (!seahorse_key_source_import_sync (lsrc, data, &err))
-            seahorse_util_handle_error (err, "Couldn't import keys");
-        
-        /* |data| is owned by the operation, no need to free */
-        
-    } else {
-        gchar *keyserver;
+    if (!seahorse_operation_is_successful (op)) {
         g_object_get (sksrc, "key-server", &keyserver, NULL);
 
         seahorse_operation_steal_error (op, &err);
@@ -80,86 +65,15 @@ sync_export_complete (SeahorseOperation *op, SeahorseKeySource *sksrc)
 static void
 ok_clicked (GtkButton *button, SeahorseWidget *swidget)
 {
-    SeahorseKeySource *lsksrc;
-    SeahorseKeySource *sksrc;
-    SeahorseMultiOperation *mop;
-    SeahorseOperation *op;
-    GError *err = NULL;
-    gpgme_error_t gerr;
-    gchar *keyserver;
-    GSList *ks, *l;
     GList *keys;
     
     keys = (GList*)g_object_get_data (G_OBJECT (swidget), "publish-keys");
     keys = g_list_copy (keys);
     
     seahorse_widget_destroy (swidget);
-    
-    if (!keys)
-        return;
-    
-    g_assert (SEAHORSE_IS_KEY (keys->data));
-
-    /* This should be the default key source */
-    lsksrc = seahorse_context_find_key_source (SCTX_APP(), SKEY_PGP, SKEY_LOC_LOCAL);
-    g_return_if_fail (lsksrc != NULL);
-
-    mop = seahorse_multi_operation_new ();
-
-    /* And now syncing keys from the servers */
-    ks = seahorse_gconf_get_string_list (KEYSERVER_KEY);
-	ks = seahorse_server_source_purge_keyservers (ks);
-	
-    for (l = ks; l; l = g_slist_next (l)) {
-        
-        sksrc = seahorse_context_remote_key_source (SCTX_APP(), (const gchar*)(l->data));
-        g_return_if_fail (sksrc != NULL);
-        
-        op = seahorse_key_source_export (sksrc, keys, FALSE, NULL);
-        g_return_if_fail (op != NULL);
-
-        g_signal_connect (op, "done", G_CALLBACK (sync_export_complete), sksrc);
-        seahorse_multi_operation_take (mop, op);
-    }
-    seahorse_util_string_slist_free (ks);
-    
-    /* Publishing keys online */    
-    keyserver = seahorse_gconf_get_string (PUBLISH_TO_KEY);
-    if (keyserver && keyserver[0]) {
-        
-        gchar *exported;
-        gpgme_data_t data;
-
-        /* Export all the necessary keys from our local keyring */
-        exported = seahorse_op_export_text (keys, FALSE, &err);
-        if (!exported) {
-            seahorse_util_handle_error (err, _("Couldn't export keys"));
-        } else {
-            /* New GPGME data which copies original text */
-            gerr = gpgme_data_new_from_mem (&data, exported, strlen (exported), 0);
-            g_return_if_fail (GPG_IS_OK (gerr));
-
-            sksrc = seahorse_context_remote_key_source (SCTX_APP (), keyserver);
-            g_return_if_fail (sksrc != NULL);
-            
-            op = seahorse_key_source_import (sksrc, data);
-            g_return_if_fail (op != NULL);
-            
-            gpgme_data_release (data);
-            g_free (exported);
-            
-            g_signal_connect (op, "done", G_CALLBACK (sync_import_complete), sksrc);
-            seahorse_multi_operation_take (mop, op);
-        }
-        
-    }
-
+   
+    seahorse_keyserver_sync (keys);
     g_list_free (keys);
-    g_free (keyserver);
-    
-    /* Show the progress window if necessary */
-    if (!seahorse_operation_is_done (SEAHORSE_OPERATION (mop))) 
-        seahorse_progress_show (SEAHORSE_OPERATION (mop), _("Syncing keys..."), FALSE);
 }
 
 static void
@@ -216,15 +130,15 @@ unhook_notification (GtkWidget *widget, gpointer data)
  **/
 GtkWindow*
 seahorse_keyserver_sync_show (GList *keys)
-{	
-	SeahorseWidget *swidget;
+{
+    SeahorseWidget *swidget;
     GtkWindow *win;
     GtkWidget *w;
     guint n, notify_id;
     gchar *t;
     
-	swidget = seahorse_widget_new ("keyserver-sync");
-	g_return_val_if_fail (swidget != NULL, NULL);
+    swidget = seahorse_widget_new ("keyserver-sync");
+    g_return_val_if_fail (swidget != NULL, NULL);
     
     win = GTK_WINDOW (glade_xml_get_widget (swidget->xml, swidget->name));
     
@@ -245,14 +159,83 @@ seahorse_keyserver_sync_show (GList *keys)
                       GINT_TO_POINTER (notify_id));
 
     keys = g_list_copy (keys);
-    g_return_val_if_fail (SEAHORSE_IS_KEY (keys->data), win);    
+    g_return_val_if_fail (!keys || SEAHORSE_IS_KEY (keys->data), win);
     g_object_set_data_full (G_OBJECT (swidget), "publish-keys", keys, 
                             (GDestroyNotify)g_list_free);
     
-	glade_xml_signal_connect_data (swidget->xml, "ok_clicked",
-		                           G_CALLBACK (ok_clicked), swidget);
-	glade_xml_signal_connect_data (swidget->xml, "configure_clicked",
-		                           G_CALLBACK (configure_clicked), swidget);
+    glade_xml_signal_connect_data (swidget->xml, "ok_clicked",
+                                   G_CALLBACK (ok_clicked), swidget);
+    glade_xml_signal_connect_data (swidget->xml, "configure_clicked",
+                                   G_CALLBACK (configure_clicked), swidget);
 
     return win;
 }
+
+void
+seahorse_keyserver_sync (GList *keys)
+{
+    SeahorseKeySource *sksrc;
+    SeahorseKeySource *lsksrc;
+    SeahorseMultiOperation *mop;
+    SeahorseOperation *op;
+    gchar *keyserver;
+    GSList *ks, *l;
+    GList *k;
+    GSList *keyids = NULL;
+    
+    if (!keys)
+        return;
+    
+    g_assert (SEAHORSE_IS_KEY (keys->data));
+    
+    /* Build a keyid list */
+    for (k = keys; k; k = g_list_next (k)) 
+        keyids = g_slist_prepend (keyids, 
+                    GUINT_TO_POINTER (seahorse_key_get_keyid (SEAHORSE_KEY (k->data))));
+
+    mop = seahorse_multi_operation_new ();
+
+    /* And now syncing keys from the servers */
+    ks = seahorse_gconf_get_string_list (KEYSERVER_KEY);
+    ks = seahorse_server_source_purge_keyservers (ks);
+    
+    for (l = ks; l; l = g_slist_next (l)) {
+        
+        sksrc = seahorse_context_remote_key_source (SCTX_APP (), (const gchar*)(l->data));
+        g_return_if_fail (sksrc != NULL);
+        
+        lsksrc = seahorse_context_find_key_source (SCTX_APP (), 
+                        seahorse_key_source_get_ktype (sksrc), SKEY_LOC_LOCAL);
+        
+        if (lsksrc) {
+            op = seahorse_transfer_operation_new (_("Syncing keys"), sksrc, lsksrc, keyids);
+            g_return_if_fail (op != NULL);
+
+            g_signal_connect (op, "done", G_CALLBACK (sync_export_complete), sksrc);
+            seahorse_multi_operation_take (mop, op);
+        }
+    }
+    
+    seahorse_util_string_slist_free (ks);
+    
+    /* Publishing keys online */    
+    keyserver = seahorse_gconf_get_string (PUBLISH_TO_KEY);
+    if (keyserver && keyserver[0]) {
+        
+        sksrc = seahorse_context_remote_key_source (SCTX_APP (), keyserver);
+        g_return_if_fail (sksrc != NULL);
+        
+        op = seahorse_context_transfer_keys (SCTX_APP (), keys, sksrc);
+        g_return_if_fail (sksrc != NULL);
+
+        g_signal_connect (op, "done", G_CALLBACK (sync_import_complete), sksrc);
+        seahorse_multi_operation_take (mop, op);
+    }
+
+    g_slist_free (keyids);
+    g_free (keyserver);
+    
+    /* Show the progress window if necessary */
+    seahorse_progress_show (SEAHORSE_OPERATION (mop), _("Syncing keys..."), FALSE);
+    g_object_unref (mop);
+}
